(* version 1.1. (c) Copyright 2006 Pierre Corbineau, Jean-Baptiste Guillaumin *)
(* published under the GNU General Public License *)

open Sys;;
open Unix;;

module type Entity =
sig

type t

val poids : t -> int

val dist : t -> t -> int 

end

let min a b c = 
 ( let min_ab = if a < b then a else b in
    if c < min_ab then c else min_ab :int)

module Word(E:Entity):Entity with type t = E.t array =
struct

  type t = E.t array

  let dist w1 w2 =
    let hauteur = Array.length w1 in
    let largeur = Array.length w2 in
    let poids1 = Array.map E.poids w1 in
    let poids2 = Array.map E.poids w2 in
    let rangee = Array.create (1+largeur) 0 in
    let diagonale = ref 0 in  
      for i = 0 to largeur-1 do
	let result = rangee.(i) + poids2.(i) in
	  rangee.(i+1)<- result;
      done;
      for j = 0 to hauteur-1 do
	diagonale:= rangee.(0);
	let result = rangee.(0) + poids1.(j) in
	  rangee.(0)<- result;
	  for i = 0 to largeur-1 do
	    let result = min 
            (rangee.(i) + poids2.(i)) 
	    (rangee.(i+1) + poids1.(j))
	    (!diagonale + E.dist w1.(j) w2.(i)) in
	      diagonale:=rangee.(i+1);
	      rangee.(i+1)<- result;
	  done;
      done;
      rangee.(largeur)
	
  let poids w = dist w [||] 

end

module Lettre =
struct
  type t = char     
  let poids c = 2
  let dist c1 c2 =
    if c1 = c2 then 0 else
      if c1 = 'v' && c2 = 'u' then 0 else
      if c1 = 'u' && c2 = 'v' then 0 else
      if c1 = 'i' && c2 = 'j' then 0 else
      if c1 = 'j' && c2 = 'i' then 0 else
      if c1 = 'e' && c2 = '' then 1 else
      if c1 = '' && c2 = 'e' then 1 else
      if c1 = '*' || c2 = '*' then 1 else 2
    
end

module Mot = Word(Lettre)

module Paragraphe =  Word(Mot)

module Texte = Word(Paragraphe)

let is_letter = function
    'a'..'z' | 'A'..'Z' | '*' | ''  -> true
  | _ -> false

let read_words file =
  let chan = open_in file in
  let word_stack = Stack.create () in
  let buf = Buffer.create 64 in
  try 
    while true do
      let c = input_char chan in
	if is_letter c then
	  Buffer.add_char buf (Char.lowercase c)
	else
	  begin
	    if Buffer.length buf > 0 then
	      Stack.push (Buffer.contents buf) word_stack;
	    Buffer.clear buf
	  end
    done;
    assert false
  with End_of_file ->
    let _ = close_in chan in
    let n_words = Stack.length word_stack in
    let text = Array.create n_words [||] in
      for i=(n_words-1) downto 0 do
	let w = Stack.pop word_stack in
	let wt=Array.init (String.length w) (fun n -> w.[n]) in
	  text.(i)<- wt
      done;
      assert (Stack.length word_stack = 0);
      text

let calcul file1 file2 = 
  let text1 = read_words file1 in
  let text2 = read_words file2 in
  let dist = Paragraphe.dist text1 text2 in
  let resultat = (dist/2) in
    resultat
   

let ecriture_init s =
let mode = [Unix.O_WRONLY;Unix.O_CREAT;Unix.O_TRUNC] in
let f = Unix.openfile "infile" mode 0o666 in
let n = Unix.write f s 0 (String.length s)
in print_string s ;
Unix.close f ;;

let ecriture s =
let mode = [Unix.O_WRONLY;Unix.O_CREAT;Unix.O_APPEND] in
let f = Unix.openfile "infile" mode 0o666 in
let n = Unix.write f s 0 (String.length s)
in print_string s ;
Unix.close f ;;

let main () =
  let nombre_fichiers = (Array.length Sys.argv) - 1 in
    ecriture_init ("   ");
    ecriture (string_of_int nombre_fichiers);

    let matricedist = Array.make_matrix (nombre_fichiers + 1) (nombre_fichiers + 1) 0 in

  for j = 1 to nombre_fichiers do
    ecriture ("\n");
    ecriture (Sys.argv.(j));

    for i = 1 to j do
   
      let resultat = matricedist.(j).(i) in
        ecriture ("   ");
        ecriture (string_of_int resultat);
        ecriture ".0";

    done;

    for i = j+1 to nombre_fichiers do
      let res = calcul Sys.argv.(i) Sys.argv.(j) in
        matricedist.(i).(j) <- res;
        matricedist.(j).(i) <- res;
        ecriture ("   ");
        ecriture (string_of_int res);
        ecriture ".0";

    done; 
  done;
        ecriture ("\n");

     

    exit 0;;
main ();;

